#include "prefix.h"
#include "zero.h"

void *KDevice::operator new(size_t, PDRIVER_OBJECT DriverObject, PCWSTR DeviceName, PCWSTR SymLinkName)
{ 
 // Create device object
 KDevice *Device=0;
 NTSTATUS ConstructorStatus;
 UNICODE_STRING Name;
 RtlInitUnicodeString(&Name, DeviceName);

 ConstructorStatus = IoCreateDevice(DriverObject,
                                  sizeof(KDevice)-sizeof(DEVICE_OBJECT),
                                  &Name,
                                  FILE_DEVICE_UNKNOWN,
                                  0,
                                  FALSE,
                                  (PDEVICE_OBJECT*)&Device);
 if(NT_SUCCESS(ConstructorStatus))
  Device->ConstructorStatus=ConstructorStatus;
 else
  return 0;

 if(SymLinkName)
 {
  RtlInitUnicodeString(&Device->SymLink, SymLinkName);
  Device->ConstructorStatus=IoCreateSymbolicLink(&Device->SymLink, &Name);
 }
 return Device;
}

KDevice::~KDevice()
{
 if(SymLink.Length)
  IoDeleteSymbolicLink(&SymLink);
 IoDeleteDevice(this);
}

NTSTATUS KDevice::Dispatch(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
 PIO_STACK_LOCATION IrpSp;

// KdBreakPoint();

 printk("KDevice::Dispatch()->");
 KDevice *Device=(KDevice *)DeviceObject;

 IrpSp=IoGetCurrentIrpStackLocation(Irp);

 switch(IrpSp->MajorFunction) 
 {
  case IRP_MJ_CREATE:
   return Device->Create(Irp);
  case IRP_MJ_CLOSE:
   return Device->Close(Irp);
  case IRP_MJ_READ:
   return Device->Read(Irp);
  case IRP_MJ_WRITE:
   return Device->Write(Irp);
  case IRP_MJ_QUERY_INFORMATION:
   return Device->QueryInformation(Irp);
  case IRP_MJ_SET_INFORMATION:
   return Device->SetInformation(Irp);
  case IRP_MJ_CLEANUP:
   return Device->Cleanup(Irp);
  default:
   printk("InvalidRequest %#02x\n", IrpSp->MajorFunction);
   Irp->IoStatus.Status = STATUS_INVALID_DEVICE_REQUEST;
   IoCompleteRequest(Irp, IO_NO_INCREMENT);
   return STATUS_INVALID_DEVICE_REQUEST;
 }
}

BOOLEAN KDevice::FastIoRead(IN PFILE_OBJECT FileObject,
                            IN PLARGE_INTEGER FileOffset,
                            IN ULONG Length,
                            IN BOOLEAN Wait,
                            IN ULONG LockKey,
                            OUT PVOID Buffer,
                            OUT PIO_STATUS_BLOCK IoStatus,
                            IN PDEVICE_OBJECT DeviceObject)
{
 KDevice *Device=(KDevice *)DeviceObject;
 return Device->FastIoRead(FileObject, FileOffset, Length, 
                           Wait, LockKey, Buffer, IoStatus);
}

BOOLEAN KDevice::FastIoWrite(IN PFILE_OBJECT FileObject,
                             IN PLARGE_INTEGER FileOffset,
                             IN ULONG Length,
                             IN BOOLEAN Wait,
                             IN ULONG LockKey,
                             OUT PVOID Buffer,
                             OUT PIO_STATUS_BLOCK IoStatus,
                             IN PDEVICE_OBJECT DeviceObject)
{
 KDevice *Device=(KDevice *)DeviceObject;
 return Device->FastIoWrite(FileObject, FileOffset, Length, 
                            Wait, LockKey, Buffer, IoStatus);
}
