#include "prefix.h"
#include "zero.h"

NTSTATUS KDevice::Read(PIRP Irp)
{
 NTSTATUS Status;
 PIO_STACK_LOCATION IrpSp;
 IrpSp=IoGetCurrentIrpStackLocation(Irp);

 printk("KDevice::Read()\n");

 if(!FlagOn(IrpSp->MinorFunction, IRP_MN_COMPLETE))
  Status = ReadNormal(Irp);
 else
 {
  Irp->IoStatus.Status = Status;
  Irp->IoStatus.Information = 0;
  IoCompleteRequest(Irp, IO_NO_INCREMENT);
 }
 return Status;
}

NTSTATUS KDevice::ReadNormal(PIRP Irp)
{
 NTSTATUS Status;
 PIO_STACK_LOCATION IrpSp;
 IrpSp=IoGetCurrentIrpStackLocation(Irp);
 ULONG Length=IrpSp->Parameters.Read.Length;

 printk("KDevice::ReadNormal(%#x, %lu)\n", Irp->UserBuffer, IrpSp->Parameters.Read.Length);
 __try
 {
  __try
  {
   if(Irp->RequestorMode!=KernelMode && !Irp->MdlAddress && Irp->UserBuffer)
   {
    ProbeForWrite(Irp->UserBuffer, Length, sizeof(UCHAR));
   }

   if(Irp->MdlAddress)
   {
    Irp->IoStatus.Information = 0;
    Status = STATUS_INVALID_PARAMETER;
    __leave;
   }

   if(Length == 0)
   {
    Irp->IoStatus.Information = 0;
    Status = STATUS_SUCCESS;
    __leave;
   }

   RtlZeroMemory(Irp->UserBuffer, Length);
   Irp->IoStatus.Information = Length;
   Status = STATUS_SUCCESS;
  }
  __except(EXCEPTION_EXECUTE_HANDLER)
  {
   Irp->IoStatus.Information = 0;
   Status = STATUS_INVALID_USER_BUFFER;
  }
 }
 __finally
 {
  Irp->IoStatus.Status = Status;
  IoCompleteRequest(Irp, IO_NO_INCREMENT);
 }
 return Status;
}

BOOLEAN KDevice::FastIoRead(IN PFILE_OBJECT FileObject,
                            IN PLARGE_INTEGER FileOffset,
                            IN ULONG Length,
                            IN BOOLEAN Wait,
                            IN ULONG LockKey,
                            OUT PVOID Buffer,
                            OUT PIO_STATUS_BLOCK IoStatus)
{
 NTSTATUS Status;
 printk("KDevice::FastIoRead(%#x, %lu)\n", Buffer, Length);
 __try
 {
  __try
  {
   ProbeForWrite(Buffer, Length, sizeof(UCHAR));

   if(Length == 0)
   {
    IoStatus->Information = 0;
    Status = STATUS_SUCCESS;
    __leave;
   }

   RtlZeroMemory(Buffer, Length);
   IoStatus->Information = Length;
   Status = STATUS_SUCCESS;
  }
  __except(EXCEPTION_EXECUTE_HANDLER)
  {
   IoStatus->Information = 0;
   Status = STATUS_INVALID_USER_BUFFER;
  }
 }
 __finally
 {
  IoStatus->Status = Status;
 }
 return TRUE;
}
